import sys

PY310: bool = sys.version_info >= (3, 10)


if sys.version_info >= (3, 10):
    from typing import Concatenate, ParamSpec, TypeAlias
else:
    from typing_extensions import Concatenate, ParamSpec, TypeAlias


__all__ = [
    "PY310",
    "Concatenate",
    "ParamSpec",
    "TypeAlias",
]



from typing import Any, Optional, List, Dict, Union, overload, TypedDict

from abc import ABC
from asyncio import gather

from hiku.error import GraphQLError
from hiku.schema import ExecutionResult, Schema


class GraphQLErrorObject(TypedDict):
    message: str


class GraphQLRequest(TypedDict, total=False):
    query: str
    variables: Optional[Dict[Any, Any]]
    operationName: Optional[str]


class GraphQLResponse(TypedDict, total=False):
    data: Optional[Dict[str, object]]
    errors: Optional[List[GraphQLErrorObject]]
    extensions: Optional[Dict[str, object]]


BatchedRequest = List[GraphQLRequest]
BatchedResponse = List[GraphQLResponse]

SingleOrBatchedRequest = Union[GraphQLRequest, BatchedRequest]
SingleOrBatchedResponse = Union[GraphQLResponse, BatchedResponse]


class BaseGraphQLEndpoint(ABC):
    """TODO: add doc explaining the purpose of this class over plain schema"""

    schema: Schema

    def __init__(
        self,
        schema: Schema,
        batching: bool = False,
    ):
        self.schema = schema
        self.batching = batching

    def process_result(self, result: ExecutionResult) -> GraphQLResponse:
        data: GraphQLResponse = {"data": result.data}

        if result.errors:
            data["errors"] = [{"message": e.message} for e in result.errors]

        return data


class BaseSyncGraphQLEndpoint(BaseGraphQLEndpoint):
    def dispatch(
        self, data: GraphQLRequest, context: Optional[Dict] = None
    ) -> GraphQLResponse:
        """
        Dispatch graphql request to graph

        Example:

        .. code-block:: python

            result = endpoint.dispatch({"query": "{ hello }"})

        :param dict data:
            {"query": str, "variables": dict, "operationName": str}
        :param dict context: context for operation
        :return: :py:class:`dict` graphql response: data or errors
        """
        assert "query" in data, "query is required"
        result = self.schema.execute_sync(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
        )
        return self.process_result(result)


class BaseAsyncGraphQLEndpoint(BaseGraphQLEndpoint):
    async def dispatch(
        self, data: GraphQLRequest, context: Optional[Dict] = None
    ) -> GraphQLResponse:
        """Dispatch graphql request to graph

        Example:

        .. code-block:: python

            result = await endpoint.dispatch({"query": "{ hello }"})

        :param dict data:
            {"query": str, "variables": dict, "operationName": str}
        :param dict context: context for operation
        :return: :py:class:`dict` graphql response: data or errors
        """
        assert "query" in data, "query is required"
        result = await self.schema.execute(
            query=data["query"],
            variables=data.get("variables"),
            operation_name=data.get("operationName"),
            context=context,
        )
        return self.process_result(result)


class GraphQLEndpoint(BaseSyncGraphQLEndpoint):
    @overload
    def dispatch(
        self, data: GraphQLRequest, context: Optional[Dict] = None
    ) -> GraphQLResponse: ...

    @overload
    def dispatch(
        self, data: BatchedRequest, context: Optional[Dict] = None
    ) -> BatchedResponse: ...

    def dispatch(
        self, data: SingleOrBatchedRequest, context: Optional[Dict] = None
    ) -> SingleOrBatchedResponse:
        if isinstance(data, list):
            if not self.batching:
                raise GraphQLError("Batching is not supported")

            return [
                super(GraphQLEndpoint, self).dispatch(item, context)
                for item in data
            ]
        else:
            return super(GraphQLEndpoint, self).dispatch(data, context)


class BatchGraphQLEndpoint(GraphQLEndpoint):
    """For backward compatibility"""

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs["batching"] = True
        super().__init__(*args, **kwargs)


class AsyncGraphQLEndpoint(BaseAsyncGraphQLEndpoint):
    @overload
    async def dispatch(
        self, data: GraphQLRequest, context: Optional[Dict] = None
    ) -> GraphQLResponse: ...

    @overload
    async def dispatch(
        self, data: BatchedRequest, context: Optional[Dict] = None
    ) -> BatchedResponse: ...

    async def dispatch(
        self, data: SingleOrBatchedRequest, context: Optional[Dict] = None
    ) -> SingleOrBatchedResponse:
        if isinstance(data, list):
            return list(
                await gather(
                    *(
                        super(AsyncGraphQLEndpoint, self).dispatch(
                            item, context
                        )
                        for item in data
                    )
                )
            )
        else:
            return await super(AsyncGraphQLEndpoint, self).dispatch(
                data, context
            )


class AsyncBatchGraphQLEndpoint(AsyncGraphQLEndpoint):
    """For backward compatibility"""

    def __init__(self, *args: Any, **kwargs: Any):
        kwargs["batching"] = True
        super().__init__(*args, **kwargs)



import typing as t

from collections import defaultdict, deque
from contextlib import contextmanager
from collections.abc import Sequence

from hiku.directives import Directive
from hiku.graph import Graph, Interface
from hiku.query import FieldOrLink, Link, Field, Fragment, Node, QueryVisitor
from hiku.types import (
    InterfaceRefMeta,
    OptionalMeta,
    Record,
    RecordMeta,
    RefMeta,
    RefMetaTypes,
    SequenceMeta,
    TypeRefMeta,
    Types,
    UnionRefMeta,
    get_type,
)


def get_ref_type(
    types: Types, type_: t.Union[t.Type[Record], RefMetaTypes], name: str
) -> RefMetaTypes:
    if isinstance(type_, RecordMeta):
        type_ = type_.__field_types__[name]
        if isinstance(type_, TypeRefMeta):
            return type_

        return get_ref_type(types, type_, name)
    elif isinstance(type_, OptionalMeta):
        assert isinstance(type_.__type__, RefMeta), type_.__type__
        return type_.__type__
    elif isinstance(type_, UnionRefMeta):
        return type_
    elif isinstance(type_, InterfaceRefMeta):
        return type_
    elif isinstance(type_, TypeRefMeta):
        type_ = get_type(types, type_)
        return get_ref_type(types, type_, name)
    elif isinstance(type_, SequenceMeta):
        type_ref = type_.__item_type__
        if isinstance(type_ref, OptionalMeta):
            type_ref = type_ref.__type__
        assert isinstance(type_ref, RefMeta), type_ref
        return type_ref
    else:
        raise AssertionError(repr(type_))


def is_match_type(
    type_: t.Union[t.Type[Record], RefMetaTypes], fragment: Fragment
) -> bool:
    if fragment.type_name is None:
        return True

    if isinstance(type_, RecordMeta):
        return False

    return fragment.type_name == type_.__type_name__


class QueryMerger(QueryVisitor):
    def __init__(self, graph: Graph):
        self.graph = graph
        self._types = graph.__types__
        self._type: t.Deque[t.Union[t.Type[Record], RefMetaTypes]] = deque(
            [self._types["__root__"]]
        )

    def merge(self, query: Node) -> Node:
        """Merge query node and return new node with merged fields, links
        and fragments.

        Must be called after the query is validated since we assume that
        no two fields with the same name+alias are present in same node.
        """
        return self.visit(query)

    @property
    def parent_type(
        self,
    ) -> t.Union[t.Type[Record], RefMetaTypes]:
        return self._type[-1]

    @contextmanager
    def _with_type_info(self, obj: Link) -> t.Iterator[None]:
        ref_type = get_ref_type(self._types, self._type[-1], obj.name)

        self._type.append(ref_type)
        yield
        self._type.pop()

    def visit_node(self, node: Node) -> Node:
        return self._merge_nodes([node])

    def visit_link(self, obj: Link) -> t.Any:
        with self._with_type_info(obj):
            return super().visit_link(obj)

    def _collect_fields(
        self,
        node: Node,
        fields: t.Dict[str, Field],
        links: t.Dict[str, t.List[Link]],
        fragments: t.List[Fragment],
    ) -> None:
        """Collect fields, links and fragments from the node.

        Fields collected in the fields dict. It is safe to assume that after
        validation, there is no possibility of having two fields with the same
        name+alias in the same node, hence we can use alias(or name) as a key.
        """
        for field in node.fields:
            name = field.alias or field.name

            if isinstance(field, Field):
                if name not in fields:
                    fields[name] = field
            elif isinstance(field, Link):
                links.setdefault(name, []).append(field)

        fragments_to_process: t.List[Fragment] = []
        for fr in node.fragments:
            if is_match_type(self.parent_type, fr):
                self._expand_fragment(fr, fields, links, fragments_to_process)
            else:
                fragments_to_process.append(fr)

        self._merge_fragments(
            fragments_to_process,
            fields,
            links,
            fragments,
        )

    def _collect_interface_fields(
        self,
        fragment: Fragment,
        interface: Interface,
        fields: t.Dict[str, Field],
        links: t.Dict[str, t.List[Link]],
    ) -> t.Optional[Fragment]:
        """For interface fragment, we collect/extract fields that are
        declared in the interface. The rest of the fields are left in
        the fragment.
        """
        fragment_fields: t.List[FieldOrLink] = []

        for field in fragment.node.fields:
            name = field.alias or field.name
            is_interface_field = name in interface.fields_map
            if isinstance(field, Field):
                if is_interface_field:
                    if name not in fields:
                        fields[name] = field
                else:
                    fragment_fields.append(field)
            elif isinstance(field, Link):
                if is_interface_field:
                    links.setdefault(name, []).append(field)
                else:
                    fragment_fields.append(field)

        if fragment_fields:
            return fragment.copy(node=Node(fragment_fields))

        return None

    def _merge_fragments(
        self,
        fragments_to_process: t.List[Fragment],
        fields: t.Dict[str, Field],
        links: t.Dict[str, t.List[Link]],
        fragments: t.List[Fragment],
    ) -> None:
        """Merge fragments. Depending on the type of the parent type,
        the fragments are merged differently.

        For example for interfaces/unions we apply diffrent logic when
        collecting fields and links.
        """
        fragments_by_type: t.DefaultDict[str, t.List[Fragment]] = defaultdict(
            list
        )

        for fragment in fragments_to_process:
            if fragment.type_name is None or (
                isinstance(self.parent_type, TypeRefMeta)
                and fragment.type_name == self.parent_type.__type_name__
            ):
                self._collect_fields(fragment.node, fields, links, fragments)
            elif isinstance(self.parent_type, InterfaceRefMeta):
                assert fragment.type_name is not None
                new_fragment = self._collect_interface_fields(
                    fragment,
                    self.graph.interfaces_map[self.parent_type.__type_name__],
                    fields,
                    links,
                )
                if new_fragment is not None:
                    fragments_by_type[fragment.type_name].append(new_fragment)
            else:
                # for the rest of the cases(including unions), we just collect
                # fragments and then merge them by type reducing duplicates
                fragments_by_type[fragment.type_name].append(fragment)

        for frs in fragments_by_type.values():
            fragments.append(self._merge_same_type_fragments(frs))

    def _merge_same_type_fragments(
        self, fragments: t.List[Fragment]
    ) -> Fragment:
        return fragments[0].copy(
            node=self._merge_nodes([fr.node for fr in fragments]),
        )

    def _expand_fragment(
        self,
        fragment: Fragment,
        fields: t.Dict[str, Field],
        links: t.Dict[str, t.List[Link]],
        fragments: t.List[Fragment],
    ) -> None:
        """Given a fragment, expand it and collect fields, links and fragments.
        Fragment is disposed.
        Example, expand QueryFragment:
        Given:
            query { ...QueryFragment }
            fragment QueryFragment on Query { id ... on Query { name } }
        Result:
            query { id name }
        """
        for field in fragment.node.fields:
            name = field.alias or field.name

            if isinstance(field, Field):
                if name not in fields:
                    fields[name] = field
            elif isinstance(field, Link):
                links.setdefault(name, []).append(field)

        for fr in fragment.node.fragments:
            fragments.append(fr)

    def _merge_nodes(self, nodes: t.List[Node]) -> Node:
        """Collect fields from multiple nodes and return new node with them"""
        assert isinstance(nodes, Sequence), type(nodes)
        ordered = any(n.ordered for n in nodes)

        fields_map: t.Dict[str, Field] = {}
        links_map: t.Dict[str, t.List[Link]] = {}
        fragments: t.List[Fragment] = []

        for node in nodes:
            self._collect_fields(node, fields_map, links_map, fragments)

        fields: t.List[FieldOrLink] = []
        for field in fields_map.values():
            fields.append(field)

        for links in links_map.values():
            fields.append(self._merge_links(links))

        return Node(fields=fields, fragments=fragments, ordered=ordered)

    def _merge_links(self, links: t.List[Link]) -> Link:
        """Recursively merge link node fields and return new link"""
        link = links[0]

        directives: t.List[Directive] = []
        for link in links:
            directives.extend(link.directives)

        with self._with_type_info(link):
            return link.copy(
                node=self._merge_nodes([link.node for link in links]),
                directives=tuple(directives),
            )



from typing import (
    Callable,
    Iterable,
    Any,
    List,
    Iterator,
    Tuple,
)

import sqlalchemy
from sqlalchemy.sql import Select
from sqlalchemy import any_
from sqlalchemy.sql.elements import BinaryExpression

from . import sqlalchemy as _sa
from ..engine import Context
from ..query import Field

# We are limiting fetch size to reduce CPU usage and avoid event-loop blocking
FETCH_SIZE = 100


def _uniq_fields(fields: List[Field]) -> Iterator[Field]:
    visited = set()
    for f in fields:
        if f.name not in visited:
            visited.add(f.name)
            yield f


class FieldsQuery(_sa.FieldsQuery):
    def in_impl(
        self, column: sqlalchemy.Column, values: Iterable
    ) -> BinaryExpression:
        return column == any_(values)

    def select_expr(
        self, fields_: List[Field], ids: Iterable
    ) -> Tuple[Select, Callable]:
        result_columns = [self.from_clause.c[f.name] for f in fields_]
        # aiopg requires unique columns to be passed to select,
        # otherwise it will raise an error
        query_columns = [
            column
            for f in _uniq_fields(fields_)
            if (column := self.from_clause.c[f.name]) != self.primary_key
        ]

        expr = (
            sqlalchemy.select(
                *_sa._process_select_params([self.primary_key] + query_columns)
            )
            .select_from(self.from_clause)
            .where(self.in_impl(self.primary_key, ids))
        )

        def result_proc(rows: List[_sa.Row]) -> List:
            rows_map = {
                row[self.primary_key]: [row[c] for c in result_columns]
                for row in map(_sa._process_result_row, rows)
            }

            nulls = [None for _ in fields_]
            return [rows_map.get(id_, nulls) for id_ in ids]

        return expr, result_proc

    async def __call__(
        self, ctx: Context, fields_: List[Field], ids: Iterable
    ) -> List:
        if not ids:
            return []

        expr, result_proc = self.select_expr(fields_, ids)

        sa_engine = ctx[self.engine_key]
        async with sa_engine.acquire() as connection:
            res = await connection.execute(expr)
            rows = []
            while True:
                bucket = await res.fetchmany(FETCH_SIZE)
                if bucket:
                    rows.extend(bucket)
                else:
                    break

        return result_proc(rows)


class LinkQuery(_sa.LinkQuery):
    def in_impl(
        self, column: sqlalchemy.Column, values: Iterable
    ) -> BinaryExpression:
        return column == any_(values)

    async def __call__(
        self, result_proc: Callable, ctx: Context, ids: Iterable
    ) -> Any:
        expr = self.select_expr(ids)
        if expr is None:
            pairs = []
        else:
            sa_engine = ctx[self.engine_key]
            async with sa_engine.acquire() as connection:
                res = await connection.execute(expr)
                rows = []
                while True:
                    bucket = await res.fetchmany(FETCH_SIZE)
                    if bucket:
                        rows.extend(bucket)
                    else:
                        break
            pairs = [(r.from_column, r.to_column) for r in rows]
        return result_proc(pairs, ids)



"""
hiku.expr.core
~~~~~~~~~~~~~~

Expression building blocks

"""

import typing as t

from functools import wraps
from itertools import chain
from collections import namedtuple

from ..compat import ParamSpec

from ..query import Node as QueryNode, Link, Field, Base as QueryBase
from ..types import (
    Record,
    Callable,
    Any,
    GenericMeta,
)

from .nodes import Symbol, Tuple, List, Keyword, Dict, Node


THIS = "this"

_Func = namedtuple("_Func", "expr, args")


class DotHandler:
    def __init__(self, obj: t.Union[Symbol, Tuple]) -> None:
        self.obj = obj

    def __repr__(self) -> str:
        return repr(self.obj)

    def __getattr__(self, name: str) -> "DotHandler":
        return DotHandler(Tuple([Symbol("get"), self.obj, Symbol(name)]))


# for backward compatibility
_DotHandler = DotHandler


class _S:
    def __getattr__(self, name: str) -> DotHandler:
        return DotHandler(Symbol(name))


#: Helper object to represent symbols in expressions. ``S.foo.bar`` in
#: expressions is equivalent to ``foo.bar`` in the regular Python.
S = _S()


def _to_expr(
    obj: t.Union[DotHandler, _Func, t.List, t.Dict, Node],
    fn_reg: t.Set[t.Callable],
) -> Node:
    if isinstance(obj, DotHandler):
        return obj.obj
    elif isinstance(obj, _Func):
        fn_reg.add(obj.expr)
        values: t.List[Node] = [Symbol(obj.expr.__def_name__)]
        values.extend(_to_expr(arg, fn_reg) for arg in obj.args)
        return Tuple(values)
    elif isinstance(obj, list):
        return List([_to_expr(v, fn_reg) for v in obj])
    elif isinstance(obj, dict):
        values = list(
            chain.from_iterable(
                (Keyword(k), _to_expr(v, fn_reg)) for k, v in obj.items()
            )
        )
        return Dict(values)
    else:
        return obj


def to_expr(
    obj: t.Union[DotHandler, _Func]
) -> t.Tuple[Node, t.Tuple[t.Callable, ...]]:
    functions: t.Set[t.Callable] = set([])
    node = _to_expr(obj, functions)
    return node, tuple(functions)


def _query_to_types(
    obj: QueryBase,
) -> t.Union[t.Type[Any], t.Type[Record]]:
    if isinstance(obj, QueryNode):
        return Record[[(f.name, _query_to_types(f)) for f in obj.fields]]
    elif isinstance(obj, Link):
        return _query_to_types(obj.node)
    elif isinstance(obj, Field):
        return Any
    else:
        raise TypeError(type(obj))


T = t.TypeVar("T")
P = ParamSpec("P")


def define(
    *types: GenericMeta, **kwargs: str
) -> t.Callable[[t.Callable[P, T]], t.Callable[P, _Func]]:
    """Annotates function arguments with types.

    These annotations are used to type-check expressions and to analyze,
    which data is used from provided arguments.

    Example:

    .. code-block:: python

        @define(Record[{'id': Integer, 'name': String}])
        def image_url(image):
            return 'http://example.com/{id}-{name}'.format(id=image['id'],
                                                           name=image['name'])

    Here ``image_url`` function accepts an object as argument, and is using two
    of it's fields: ``id`` field of type ``Integer`` and ``name`` field of type
    ``String``. Hiku will check that this function will be used only with
    objects having at least such two fields.

    This annotation also gives ability for Hiku to build a query for low-level
    graph.
    """

    def decorator(fn: t.Callable[P, T]) -> t.Callable[P, _Func]:
        _name = kwargs.pop("_name", None)
        assert not kwargs, repr(kwargs)

        name = _name or "{}/{}_{}".format(fn.__module__, fn.__name__, id(fn))

        @wraps(fn)
        def expr(*args: P.args, **kw: P.kwargs) -> _Func:
            return _Func(expr, args)

        expr.__def_name__ = name  # type: ignore[attr-defined]
        expr.__def_body__ = fn  # type: ignore[attr-defined]
        expr.__def_type__ = Callable[types]  # type: ignore[attr-defined]
        return expr

    return decorator


@define(Any, Any, Any, _name="each")
def each(var: DotHandler, col: DotHandler, expr: DotHandler) -> None:
    """Returns a list of the results of the expression evaluation for every
    item of the sequence provided.

    Example:

    .. code-block:: python

        each(S.x, S.collection, S.x.name)

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        [x.name for x in collection]

    """


@define(Any, Any, Any, _name="if")
def if_(test: t.Any, then: t.Any, else_: t.Any) -> None:
    """Checks condition and continues to evaluate one of the two expressions
    provided.

    Example:

    .. code-block:: python

        if_(S.value, 'truish', 'falsish')

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        if value:
            return 'truish'
        else:
            return 'falsish'

    """


@define(Any, Any, Any, _name="if_some")
def if_some(bind: t.List, then: _Func, else_: t.Any) -> None:
    """Used to unpack values with ``Optional`` types and using them safely in
    expressions.

    Example:

    .. code-block:: python

        if_some([S.img, S.this.image],
                image_url(S.img),
                'http://example.com/no-photo.jpg')

    Equivalent in the regular Python (only for reference):

    .. code-block:: python

        if this.image is not None:
            img = this.image
            return image_url(img)
        else:
            return 'http://example.com/no-photo.jpg'

    If ``S.this.image`` has a type ``Optional[TypeRef['Image']]``, ``S.img``
    variable will have a type ``TypeRef['Image']`` and it will be available only
    in "then"-expression, which will be evaluated only if ``S.this.image``
    wouldn't be ``None``. Otherwise "else"-expression will be
    evaluated.
    """



import typing as t
from collections import abc as collections_abc
from contextlib import contextmanager
from dataclasses import dataclass

from hiku.graph import (
    Field,
    Graph,
    GraphVisitor,
    Input,
    Interface,
    Link,
    LinkType,
    Node,
    Nothing,
    Option,
    Root,
    Union,
)
from hiku.query import Field as QueryField
from hiku.query import FieldBase, Fragment
from hiku.query import Link as QueryLink
from hiku.query import Node as QueryNode
from hiku.query import QueryVisitor
from hiku.scalar import ScalarMeta

from ..types import (
    AbstractTypeVisitor,
    AnyMeta,
    BooleanMeta,
    EnumRefMeta,
    FloatMeta,
    GenericMeta,
    IDMeta,
    InputRefMeta,
    IntegerMeta,
    MappingMeta,
    OptionalMeta,
    Record,
    RecordMeta,
    SequenceMeta,
    StringMeta,
    TypeRefMeta,
)
from .errors import Errors

_undefined = object()


OptionValue = t.Union[
    int, str, float, bool, collections_abc.Sequence, collections_abc.Mapping
]


@dataclass
class Path:
    parent: t.Optional["Path"]
    path: t.Union[str, t.Tuple[str, str]]

    def add(self, node_name: str, obj_name: str) -> "Path":
        return Path(self, (node_name, obj_name))

    def pop(self) -> "Path":
        assert self.parent is not None, "Can not pop root path"
        return self.parent

    def __hash__(self) -> int:
        return hash(self.parent) + hash(self.path)


class _AssumeRecord(AbstractTypeVisitor):
    def __init__(
        self, data_types: t.Dict[str, t.Type[Record]], _nested: bool = False
    ):
        self._data_types = data_types
        self._nested = _nested

    def _get_nested(self) -> "_AssumeRecord":
        return _AssumeRecord(self._data_types, _nested=True)

    def visit(self, obj: t.Any) -> t.Any:
        if obj is not None:
            return obj.accept(self)

    def _false(self, obj: t.Any) -> None:
        pass

    visit_any = _false
    visit_boolean = _false
    visit_string = _false
    visit_id = _false
    visit_integer = _false
    visit_float = _false
    visit_mapping = _false
    visit_callable = _false
    visit_unionref = _false
    visit_interfaceref = _false
    visit_inputref = _false
    visit_enumref = _false
    visit_scalar = _false

    def visit_optional(self, obj: OptionalMeta) -> t.Optional[t.OrderedDict]:
        if not self._nested:
            return self._get_nested().visit(obj.__type__)
        return None

    def visit_sequence(self, obj: SequenceMeta) -> t.Optional[t.OrderedDict]:
        if not self._nested:
            return self._get_nested().visit(obj.__item_type__)
        return None

    def visit_record(self, obj: RecordMeta) -> t.OrderedDict:
        # return fields alongside type definitions
        return obj.__field_types__

    def visit_typeref(self, obj: TypeRefMeta) -> t.OrderedDict:
        return self.visit(self._data_types[obj.__type_name__])


class _AssumeField(GraphVisitor):
    def __init__(self, node: Node, errors: Errors) -> None:
        self.node = node
        self.errors = errors

    def visit_field(self, obj: Field) -> bool:
        return True

    def visit_link(self, obj: Link) -> bool:
        self.errors.report(
            'Trying to query "{}.{}" link as it was a field'.format(
                self.node.name or "root", obj.name
            )
        )
        return False

    def visit_node(self, obj: Node) -> bool:
        assert (
            self.node.name is None
        ), "Nested node can be only in the root node"
        self.errors.report(
            'Trying to query "{}" node as it was a field'.format(obj.name)
        )
        return False

    def visit_root(self, obj: Root) -> t.NoReturn:
        raise AssertionError("Root node is not expected here")


class _OptionError(TypeError):
    def __init__(self, description: str, field: t.Optional[str] = None) -> None:
        self.description = description
        self.field = field
        super(_OptionError, self).__init__(description)


class _OptionTypeError(_OptionError):
    def __init__(self, value: OptionValue, expected: GenericMeta) -> None:
        type_name = type(value).__name__
        if value is Nothing:
            type_name = "Nothing"
        description = '"{}" instead of {!r}'.format(type_name, expected)
        super(_OptionTypeError, self).__init__(description)


class _OptionTypeValidator:
    def __init__(
        self,
        data_types: t.Dict[str, t.Type[Record]],
        inputs_map: t.Dict[str, Input],
        value: OptionValue,
    ) -> None:
        self._data_types = data_types
        self._inputs_map = inputs_map
        self._value = [value]

    @property
    def value(self) -> OptionValue:
        return self._value[-1]

    @contextmanager
    def push_value(self, value: OptionValue) -> t.Iterator[None]:
        self._value.append(value)
        try:
            yield
        finally:
            self._value.pop()

    def visit(self, type_: GenericMeta) -> None:
        type_.accept(self)  # type: ignore

    def visit_any(self, type_: AnyMeta) -> None:
        pass

    def visit_boolean(self, type_: BooleanMeta) -> None:
        if not isinstance(self.value, bool):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_string(self, type_: StringMeta) -> None:
        if not isinstance(self.value, str):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_id(self, type_: IDMeta) -> None:
        if not isinstance(self.value, str):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_integer(self, type_: IntegerMeta) -> None:
        if not isinstance(self.value, int):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_float(self, type_: FloatMeta) -> None:
        if not isinstance(self.value, (int, float)):
            raise _OptionTypeError(self.value, type_)
        return None

    def visit_optional(self, type_: OptionalMeta) -> None:
        if self.value is not None:
            self.visit(type_.__type__)
        return None

    def visit_sequence(self, type_: SequenceMeta) -> None:
        if not isinstance(self.value, collections_abc.Sequence):
            raise _OptionTypeError(self.value, type_)
        for item in self.value:
            with self.push_value(item):
                self.visit(type_.__item_type__)

        return None

    def visit_mapping(self, type_: MappingMeta) -> None:
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)
        for key, value in self.value.items():
            with self.push_value(key):
                self.visit(type_.__key_type__)
            with self.push_value(value):
                self.visit(type_.__value_type__)

        return None

    def visit_record(self, type_: RecordMeta) -> None:
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)

        unknown = set(self.value).difference(type_.__field_types__)
        if unknown:
            fields = ", ".join(sorted(map(repr, unknown)))
            raise _OptionError("unknown fields: {}".format(fields))

        missing = set(type_.__field_types__).difference(self.value)
        if missing:
            fields = ", ".join(sorted(missing))
            raise _OptionError("missing fields: {}".format(fields))

        for key, value_type in type_.__field_types__.items():
            with self.push_value(self.value[key]):
                self.visit(value_type)

        return None

    def visit_typeref(self, type_: TypeRefMeta) -> None:
        assert (
            type_.__type_name__ in self._data_types
        ), f'"{type_.__type_name__}" type is not present in graph data_types'
        self.visit(self._data_types[type_.__type_name__])

    # TODO: add scalar validation errors
    def visit_scalar(self, type_: ScalarMeta) -> None:
        pass

    # TODO: add enum validation errors
    def visit_enumref(self, type_: EnumRefMeta) -> None:
        pass

    def visit_inputref(self, type_: InputRefMeta) -> None:
        if not isinstance(self.value, collections_abc.Mapping):
            raise _OptionTypeError(self.value, type_)

        input_type = self._inputs_map[type_.__type_name__]

        # check for unknown fields in the input
        unknown = set(self.value).difference(input_type.arguments_map)
        if unknown:
            fields = ", ".join(sorted(map(repr, unknown)))
            raise _OptionError("unknown fields: {}".format(fields))

        non_default_args = {
            k: v
            for k, v in input_type.arguments_map.items()
            # option with default value is not required
            if v.default is Nothing
            # option with optional type is not required
            and not isinstance(v.type, OptionalMeta)
        }
        missing = set(non_default_args).difference(self.value)
        if missing:
            fields = ", ".join(sorted(missing))
            raise _OptionError("missing fields: {}".format(fields))

        for key, value_type in input_type.arguments_map.items():
            value = self.value.get(key, value_type.default)
            if isinstance(value_type.type, OptionalMeta) and value is Nothing:
                # if value is not specified for optional argument,
                # we can skip validation
                continue
            if value_type.type is None:
                raise _OptionError("type is not specified", key)
            with self.push_value(value):
                try:
                    self.visit(value_type.type)
                except _OptionError as err:
                    err.field = key
                    raise err

        return None


class _ValidateOptions(GraphVisitor):
    """
    Validate options for fields and links.

    :param data_types: Mapping of data types.
    :param options: Options to validate.
    :param for_: Path to the field or link. Used in error messages.
    :param errors: Errors container.
    """

    def __init__(
        self,
        data_types: t.Dict[str, t.Type[Record]],
        inputs_map: t.Dict[str, Input],
        options: t.Optional[t.Dict],
        for_: t.Tuple[t.Any, ...],
        errors: Errors,
    ) -> None:
        self._data_types = data_types
        self._inputs_map = inputs_map
        self.options = options
        self.for_ = for_
        self.errors = errors
        self._options = options or {}

    def visit_link(self, obj: Link) -> None:
        super(_ValidateOptions, self).visit_link(obj)
        unknown = set(self._options).difference(obj.options_map)
        if unknown:
            node, field = self.for_
            self.errors.report(
                'Unknown options for "{}.{}": {}'.format(
                    node, field, ", ".join(unknown)
                )
            )

        return None

    visit_field = visit_link  # type: ignore

    def visit_option(self, obj: Option) -> None:
        value = self._options.get(obj.name, obj.default)
        if value is Nothing:
            node, field = self.for_

            if obj.type and isinstance(obj.type, InputRefMeta):
                error = 'Required option "{}.{}:{}" of type "{}" is not specified'.format(  # noqa: E501
                    node, field, obj.name, obj.type.__type_name__
                )
            else:
                error = 'Required option "{}.{}:{}" is not specified'.format(
                    node, field, obj.name
                )
            self.errors.report(error)

        elif obj.type is not None:
            try:
                _OptionTypeValidator(
                    self._data_types, self._inputs_map, value
                ).visit(obj.type)
            except _OptionError as err:
                node, field = self.for_
                if err.field is not None:
                    error = 'Invalid value for option "{}.{}:{}.{}", {}'.format(
                        node, field, obj.name, err.field, err.description
                    )
                else:
                    error = 'Invalid value for option "{}.{}:{}", {}'.format(
                        node, field, obj.name, err.description
                    )
                self.errors.report(error)

        return None

    def visit_node(self, obj: Node) -> None:
        assert self.options is None, "Node can not have options"

    def visit_root(self, obj: Root) -> t.NoReturn:
        raise AssertionError("Root node is not expected here")


class _RecordFieldsValidator(QueryVisitor):
    def __init__(
        self,
        data_types: t.Dict[str, t.Type[Record]],
        field_types: t.OrderedDict,
        errors: Errors,
    ) -> None:
        self._data_types = data_types
        self._field_types = field_types
        self._errors = errors

    def visit_field(self, obj: QueryField) -> None:
        if obj.name == "__typename":
            return

        if obj.name not in self._field_types:
            self._errors.report('Unknown field name "{}"'.format(obj.name))
        elif obj.options is not None:
            self._errors.report("Options are not expected")
        elif _AssumeRecord(self._data_types).visit(self._field_types[obj.name]):
            self._errors.report(
                'Trying to query "{}" link as it was a field'.format(obj.name)
            )

    def visit_link(self, obj: QueryLink) -> None:
        field_types = _AssumeRecord(self._data_types).visit(
            self._field_types[obj.name]
        )
        if field_types is not None:
            fields_validator = _RecordFieldsValidator(
                self._data_types, field_types, self._errors
            )
            for field in obj.node.fields:
                fields_validator.visit(field)
        else:
            self._errors.report('"{}" is not a link'.format(obj.name))

    def visit_node(self, obj: QueryNode) -> None:
        raise AssertionError("Node is not expected here")


def _field_eq(a: FieldBase, b: FieldBase) -> bool:
    return a.name == b.name and a.options == b.options


class DefaultQueryValidator(QueryVisitor):
    """
    Validate query against graph.

    Query must not contain __typename field.

    :param graph: Graph to validate against.
    """

    def __init__(self, graph: Graph):
        self.graph = graph
        self._type = [graph.root]
        self._path: Path = Path(None, "root")
        self.errors = Errors()
        self._visited_fields: t.Dict[Path, t.Dict] = {}

    def visit_field(self, obj: QueryField) -> None:
        node = self._type[-1]
        field = node.fields_map.get(obj.name)
        if field is not None:
            is_field = _AssumeField(node, self.errors).visit(field)
            if is_field:
                for_ = (node.name or "root", obj.name)
                _ValidateOptions(
                    self.graph.data_types,
                    self.graph.inputs_map,
                    obj.options,
                    for_,
                    self.errors,
                ).visit(field)
        else:
            self.errors.report(
                'Field "{}" is not implemented in the "{}" node'.format(
                    obj.name, node.name or "root"
                )
            )

    def visit_link(self, obj: QueryLink) -> None:
        node = self._type[-1]
        graph_obj = node.fields_map.get(obj.name, _undefined)
        if isinstance(graph_obj, Field):
            for_ = (node.name or "root", obj.name)
            _ValidateOptions(
                self.graph.data_types,
                self.graph.inputs_map,
                obj.options,
                for_,
                self.errors,
            ).visit(graph_obj)

            field_types = _AssumeRecord(self.graph.data_types).visit(
                graph_obj.type
            )
            if field_types is not None:
                fields_validator = _RecordFieldsValidator(
                    self.graph.data_types, field_types, self.errors
                )
                for field in obj.node.fields:
                    fields_validator.visit(field)
            else:
                self.errors.report(
                    'Trying to query "{}.{}" simple field '
                    "as node".format(node.name or "root", obj.name)
                )

        elif isinstance(graph_obj, Link):
            if graph_obj.type_info.type_enum is LinkType.UNION:
                linked_node = self.graph.unions_map[graph_obj.node]
            elif graph_obj.type_info.type_enum is LinkType.INTERFACE:
                linked_node = self.graph.interfaces_map[graph_obj.node]
            else:
                linked_node = self.graph.nodes_map[graph_obj.node]

            for_ = (node.name or "root", obj.name)
            _ValidateOptions(
                self.graph.data_types,
                self.graph.inputs_map,
                obj.options,
                for_,
                self.errors,
            ).visit(graph_obj)

            self._type.append(linked_node)
            self._path = self._path.add(*for_)
            try:
                self.visit(obj.node)
            finally:
                self._type.pop()
                self._path = self._path.pop()

        elif graph_obj is _undefined:
            self.errors.report(
                'Link "{}" is not implemented in the "{}" node'.format(
                    obj.name, node.name or "root"
                )
            )
        else:
            raise TypeError(repr(graph_obj))

    def visit_fragment(self, obj: Fragment) -> t.Any:
        graph_node = None
        if isinstance(self._type[-1], Union):
            if (
                obj.type_name is not None
                and obj.type_name in self._type[-1].types
            ):
                graph_node = self.graph.nodes_map[obj.type_name]
            elif (
                obj.type_name is not None
                and obj.type_name != self._type[-1].name
            ):
                if obj.type_name not in self.graph.nodes_map:
                    self.errors.report(
                        "Fragment on unknown type '{}'".format(obj.type_name)
                    )
                    return
                else:
                    self.errors.report(
                        "Fragment type '{}' does not match "
                        "the union type '{}'".format(
                            obj.type_name, self._type[-1].name
                        )
                    )
        elif isinstance(self._type[-1], Interface):
            if (
                obj.type_name is not None
                and obj.type_name
                in self.graph.interfaces_types[self._type[-1].name]
            ):
                graph_node = self.graph.nodes_map[obj.type_name]
            elif (
                obj.type_name is not None
                and obj.type_name != self._type[-1].name
            ):
                if obj.type_name not in self.graph.nodes_map:
                    self.errors.report(
                        "Fragment on unknown type '{}'".format(obj.type_name)
                    )
                    return
                else:
                    self.errors.report(
                        "Fragment type '{}' does not match "
                        "the interface type '{}'".format(
                            obj.type_name, self._type[-1].name
                        )
                    )
        elif obj.type_name in ("Query", "Mutation") or (
            self._type[-1].name is None and obj.type_name is None
        ):
            graph_node = self.graph.root
        elif obj.type_name not in self.graph.nodes_map:
            self.errors.report(
                "Fragment on an unknown type '{}'".format(obj.type_name)
            )
            return
        else:
            graph_node = self.graph.nodes_map[obj.type_name]

        if graph_node:
            self._type.append(graph_node)

        self.visit(obj.node)

        if graph_node:
            self._type.pop()

    def visit_node(self, obj: QueryNode) -> None:
        is_union_link = isinstance(self._type[-1], Union)
        is_interface_link = isinstance(self._type[-1], Interface)

        fields: t.Dict = self._visited_fields.setdefault(self._path, {})

        for field in obj.fields:
            if field.name == "__typename":
                continue

            if is_union_link:
                union = t.cast(Union, self._type[-1])
                self.errors.report(
                    "Cannot query field '{}' on type '{}'. "
                    "Did you mean to use an inline fragment on {}?".format(
                        field.name,
                        union.name,
                        " or ".join(
                            [f"'{type_name}'" for type_name in union.types]
                        ),
                    )
                )
                continue
            elif is_interface_link:
                interface = self._type[-1]
                assert interface.name, "Interface must have a name"

                interface_types = self.graph.interfaces_types[interface.name]
                if not interface_types:
                    self.errors.report(
                        "Can not query field '{0}' on interface '{1}'. "
                        "Interface '{1}' is not implemented by any type. "
                        "Add at least one type implementing this interface.".format(  # noqa: E501
                            field.name, interface.name
                        )
                    )
                    continue

                if field.name not in interface.fields_map:
                    implementation = None
                    for impl in interface_types:
                        if field.name in self.graph.nodes_map[impl].fields_map:
                            implementation = impl
                            break

                    self.errors.report(
                        "Can not query field '{}' on type '{}'. "
                        "Did you mean to use an inline fragment on '{}'?".format(  # noqa: E501
                            field.name, interface.name, implementation
                        )
                    )
                    continue

            seen = fields.get(field.result_key)
            if seen is not None:
                if not _field_eq(field, seen):
                    node = self._type[-1].name or "root"
                    self.errors.report(
                        "Found distinct fields with the same "
                        'resulting name "{}" for the node "{}"'.format(
                            field.result_key, node
                        )
                    )
            else:
                fields[field.result_key] = field
            self.visit(field)

            if is_union_link:
                self.path.pop()

        for fr in obj.fragments:
            self.visit(fr)


def validate(graph: Graph, query: QueryNode) -> t.List[str]:
    query_validator = DefaultQueryValidator(graph)
    query_validator.visit(query)
    return query_validator.errors.list



"""
    hiku.graph
    ~~~~~~~~~~

    Graphs are defined by nodes, fields and links. Simple functions
    are used to fetch any data from any data source.

"""

import dataclasses
import typing as t

from abc import ABC, abstractmethod
from enum import Enum
from itertools import chain
from functools import reduce, cached_property
from collections import OrderedDict, defaultdict
from typing import List

from hiku.enum import BaseEnum
from .scalar import Scalar, ScalarMeta

from .types import (
    EnumRefMeta,
    InputRefMeta,
    InterfaceRef,
    InterfaceRefMeta,
    Optional,
    OptionalMeta,
    RefMeta,
    Sequence,
    SequenceMeta,
    TypeRef,
    Record,
    Any,
    GenericMeta,
    TypeRefMeta,
    TypingMeta,
    AnyMeta,
    UnionRef,
    UnionRefMeta,
)
from .utils import (
    const,
    Const,
)
from .directives import Deprecated, SchemaDirective

from .compat import TypeAlias


if t.TYPE_CHECKING:
    from .sources.graph import SubGraph
    from .sources.graph import BoundExpr

# TODO enum ???
Maybe = const("Maybe")

One = const("One")

Many = const("Many")

MaybeMany = const("MaybeMany")

#: Special constant that is used by links with :py:class:`~hiku.types.Optional`
#: type in order to indicate that there is nothing to link to
Nothing = const("Nothing")
NothingType: TypeAlias = Const


class AbstractBase(ABC):
    @abstractmethod
    def accept(self, visitor: "AbstractGraphVisitor") -> None:
        pass


class AbstractOption(AbstractBase, ABC):
    pass


class Option(AbstractOption):
    """Defines an option, used to customize results of the fields and links

    Options without default value are **required**.

    Example of a required option::

        Option('id', Integer)

    Example of an optional option::

        Option('size', Integer, default=100)

    Example with TypeRef type(ref can point either to Node or data type)::

        Option('filter', TypeRef['FilterInput'])

    Example with InputRef type(ref can point to Input)::

        Option('filter', InputRef['FilterInput'])
    """

    def __init__(
        self,
        name: str,
        type_: t.Optional[GenericMeta],
        *,
        default: t.Any = Nothing,
        description: t.Optional[str] = None,
    ):
        """
        :param name: name of the option
        :param type_: type of the option or ``None``
        :param default: default option value
        :param description: description of the option
        """
        self.name = name
        # TODO(m.kind): make type_ non-optional
        self.type = type_
        self.default = default
        self.description = description
        self.type_info = get_field_info(type_)

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.type
        )

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_option(self)


class AbstractField(AbstractBase, ABC):
    pass


R = t.TypeVar("R")

SyncAsync = t.Union[R, t.Awaitable[R]]

# (fields) -> []
RootFieldFunc = t.Callable[[t.List["Field"]], SyncAsync[List[t.Any]]]
# (fields, ids) -> [[]]
NotRootFieldFunc = t.Callable[
    [t.List["Field"], List[t.Any]], SyncAsync[List[List[t.Any]]]
]
# (ctx, fields, ids) -> [[]]
NotRootFieldFuncCtx = t.Callable[
    [t.Any, t.List["Field"], List[t.Any]], SyncAsync[List[List[t.Any]]]
]


FieldT = t.Optional[t.Union[GenericMeta, ScalarMeta]]
FieldFunc = t.Union[
    RootFieldFunc,
    NotRootFieldFunc,
    NotRootFieldFuncCtx,
    "SubGraph",
    "BoundExpr",
]


class Field(AbstractField):
    """Defines a field of the node

    Example::

        graph = Graph([
            Node('user', [
                Field('name', String, func),
            ]),
        ])

    Example with options::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)]),
            ]),
        ])

    Example with directives::

        graph = Graph([
            Root([
                Field('lorem-ipsum', String, func,
                      options=[Option('words', Integer, default=50)],
                      directives=[Deprecated('use another field')]),
            ]),
        ])

    Data loading protocol::

        # root node fields
        def func(fields) -> List[T]

        # non-root node fields
        def func(fields, ids) -> List[List[T]]

    Where:

    - ``fields`` - list of :py:class:`hiku.query.Field`
    - ``ids`` - list node identifiers

    """

    def __init__(
        self,
        name: str,
        type_: FieldT,
        func: FieldFunc,
        *,
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ):
        """
        :param str name: name of the field
        :param type_: type of the field or ``None``
        :param func: function to load field's data
        :param options: list of acceptable options
        :param description: description of the field
        :param directives: list of directives for the field
        :param deprecated: deprecation reason
        """
        if directives is None:
            directives = []

        if deprecated is not None:
            directives.append(Deprecated(deprecated))

        self.name = name
        self.type = type_
        self.func = func
        self.options = options or ()
        self.description = description
        self.directives = directives
        self.type_info = get_field_info(type_)

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, {!r})".format(
            self.__class__.__name__, self.name, self.type, self.func
        )

    @cached_property
    def options_map(self) -> OrderedDict:
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_field(self)


class AbstractLink(AbstractBase, ABC):
    pass


def get_link_type_enum(type_: TypingMeta) -> t.Tuple[Const, str]:
    if isinstance(type_, RefMeta):
        return One, type_.__type_name__
    elif isinstance(type_, OptionalMeta):
        if isinstance(type_.__type__, RefMeta):
            return Maybe, type_.__type__.__type_name__
    elif isinstance(type_, SequenceMeta):
        if isinstance(type_.__item_type__, RefMeta):
            return Many, type_.__item_type__.__type_name__
        elif isinstance(type_.__item_type__, OptionalMeta):
            if isinstance(type_.__item_type__.__type__, RefMeta):
                return MaybeMany, type_.__item_type__.__type__.__type_name__
    raise TypeError("Invalid type specified: {!r}".format(type_))


class FieldType(Enum):
    # When Field is a TypeRef to Record
    RECORD = "RECORD"
    # When Field is a InputRef to Input
    INPUT = "INPUT"
    # When Field is a EnumRef to Enum
    ENUM = "ENUM"
    # When Field is a builtin scalar (such as String, Integer, etc)
    SCALAR = "SCALAR"
    # When Field is a custom scalar (such as DateTime, etc)
    CUSTOM_SCALAR = "CUSTOM_SCALAR"


@dataclasses.dataclass
class FieldTypeInfo:
    type_name: str
    type_enum: FieldType


def get_field_info(
    type_: t.Optional[t.Union[GenericMeta, ScalarMeta]]
) -> t.Optional[FieldTypeInfo]:
    if not type_:
        return None

    if isinstance(type_, OptionalMeta):
        return get_field_info(getattr(type_, "__type__", None))

    if isinstance(type_, SequenceMeta):
        # if type not passed to Sequence[]
        return get_field_info(getattr(type_, "__item_type__", None))

    if isinstance(type_, TypeRefMeta):
        return FieldTypeInfo(
            type_.__type_name__,
            FieldType.RECORD,
        )

    if isinstance(type_, InputRefMeta):
        return FieldTypeInfo(
            type_.__type_name__,
            FieldType.INPUT,
        )
    elif isinstance(type_, EnumRefMeta):
        return FieldTypeInfo(type_.__type_name__, FieldType.ENUM)
    elif isinstance(type_, ScalarMeta):
        return FieldTypeInfo(type_.__type_name__, FieldType.CUSTOM_SCALAR)

    if isinstance(type_, GenericMeta):
        return FieldTypeInfo(type_.__name__, FieldType.SCALAR)

    return None


class LinkType(Enum):
    NODE = "NODE"
    INTERFACE = "INTERFACE"
    UNION = "UNION"


@dataclasses.dataclass
class LinkTypeInfo:
    type_name: str
    type_enum: LinkType


def get_link_type(type_: GenericMeta) -> LinkTypeInfo:
    if isinstance(type_, OptionalMeta):
        return get_link_type(type_.__type__)

    if isinstance(type_, SequenceMeta):
        return get_link_type(type_.__item_type__)

    if isinstance(type_, TypeRefMeta):
        return LinkTypeInfo(
            type_.__type_name__,
            LinkType.NODE,
        )
    elif isinstance(type_, UnionRefMeta):
        return LinkTypeInfo(type_.__type_name__, LinkType.UNION)
    elif isinstance(type_, InterfaceRefMeta):
        return LinkTypeInfo(type_.__type_name__, LinkType.INTERFACE)

    raise TypeError("Invalid link type specified: {!r}".format(type_))


def collect_interfaces_types(
    items: t.List["Node"], interfaces: t.List["Interface"]
) -> t.Dict[str, t.List[str]]:
    interfaces_types = defaultdict(list)
    for item in items:
        if item.name is not None and item.implements:
            for impl in item.implements:
                interfaces_types[impl].append(item.name)

    for i in interfaces:
        if i.name not in interfaces_types:
            interfaces_types[i.name] = []

    return dict(interfaces_types)


LT = t.TypeVar("LT", bound=t.Hashable)
LR = t.TypeVar("LR", bound=t.Optional[t.Hashable])

MaybeLink = t.Union[LR, NothingType]
RootLinkT = t.Union[
    # () -> LR
    t.Callable[[], SyncAsync[LR]],
    # (opts) -> LR
    t.Callable[[t.Any], SyncAsync[LR]],
    # (ctx, opts) -> LR
    t.Callable[[t.Any, t.Any], SyncAsync[LR]],
]

LinkT = t.Union[
    # (ids) -> []
    t.Callable[[List[LT]], SyncAsync[LR]],
    # (ids, opts) -> []
    t.Callable[[List[LT], t.Any], SyncAsync[LR]],
    # (ctx, ids) -> []
    t.Callable[[t.Any, List[LT]], SyncAsync[LR]],
    # (ctx, ids, opts) -> []
    t.Callable[[t.Any, List[LT], t.Dict], SyncAsync[LR]],
]

RootLinkOne = RootLinkT[LR]
RootLinkMaybe = RootLinkT[MaybeLink[LR]]
RootLinkMany = RootLinkT[List[LR]]  # type: ignore[type-var]

LinkOne = LinkT[LT, List[LR]]  # type: ignore[type-var]
LinkMaybe = LinkT[LT, List[MaybeLink[LR]]]  # type: ignore[type-var]
LinkMany = LinkT[LT, List[List[LR]]]  # type: ignore[type-var]

LinkFunc = t.Union[
    RootLinkOne,
    RootLinkMaybe,
    RootLinkMany,
    LinkOne,
    LinkMaybe,
    LinkMany,
]

LinkOneFunc = t.Union[RootLinkOne, LinkOne]
LinkMaybeFunc = t.Union[RootLinkMaybe, LinkMaybe]
LinkManyFunc = t.Union[RootLinkMany, LinkMany]


class Link(AbstractLink):
    """Defines a link to the node

    Example::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None),
            ]),
        ])

    Example with requirements::

        graph = Graph([
            Node('character', [...]),
            Node('actor', [
                Field('id', Integer, field_func),
                Link('characters', Sequence[TypeRef['character']],
                     link_func, requires='id'),
            ])
        ])

    Requirements are needed when link points from non-root node.

    Example with options::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)]),
            ]),
        ])

    Example with directives::

        graph = Graph([
            Node('user', [...]),
            Root([
                Link('users', Sequence[TypeRef['user']], func, requires=None,
                     options=[Option('limit', Integer, default=100)],
                     directives=[Deprecated('do not use')]),
            ]),
        ])

    Identifiers loading function protocol::

        # From root node or if requires is None:

        # ... if type is TypeRef['foo']
        def func() -> T

        # ... if type is Optional[TypeRef['foo']]
        def func() -> Union[T, Nothing]

        # ... if type is Sequence[TypeRef['foo']]
        def func() -> List[T]

        # From non-root node and requires is not None:

        # ... if type is TypeRef['foo']
        def func(ids) -> List[T]

        # ... if type is Optional[TypeRef['foo']]
        def func(ids) -> List[Union[T, Nothing]]

        # ... if type is Sequence[TypeRef['foo']]
        def func(ids) -> List[List[T]]

    See also :py:const:`hiku.graph.Nothing`.

    If link was defined with options, then link function should accept one
    additional positional argument::

        # many to many link with options
        def func(ids, options) -> List[List[T]]

    Where ``options`` is a mapping ``str: value`` of provided in the query
    options.
    """

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[TypeRef],
        func: LinkOneFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ): ...

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[t.Union[UnionRef, InterfaceRef]],
        func: LinkOneFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ): ...

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[Optional],
        func: LinkMaybeFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ): ...

    @t.overload
    def __init__(
        self,
        name: str,
        type_: t.Type[Sequence],
        func: LinkManyFunc,
        *,
        requires: t.Optional[t.Union[str, t.List[str]]],
        options: t.Optional[t.Sequence[Option]] = None,
        description: t.Optional[str] = None,
        directives: t.Optional[t.List[SchemaDirective]] = None,
        deprecated: t.Optional[str] = None,
    ): ...

    def __init__(  # type: ignore[no-untyped-def]
        self,
        name,
        type_,
        func,
        *,
        requires,
        options=None,
        description=None,
        directives=None,
        deprecated=None,
    ):
        """
        :param name: name of the link
        :param type_: type of the link
        :param func: function to load identifiers of the linked node
        :param requires: field name(s) from the current node,
            required to compute identifiers of the linked node
        :param options: list of acceptable options
        :param description: description of the link
        :param directives: list of directives for the link
        :param deprecated: deprecation reason for the link
        """
        type_enum, node = get_link_type_enum(type_)

        if directives is None:
            directives = []

        if deprecated is not None:
            directives.append(Deprecated(deprecated))

        self.name = name
        self.type = type_
        self.type_enum = type_enum
        self.node = node
        self.func = func
        self.requires = requires
        self.options = options or ()
        self.description = description
        self.directives = directives
        self.type_info = get_link_type(type_)

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.type, self.func
        )

    @cached_property
    def options_map(self) -> OrderedDict:
        return OrderedDict((op.name, op) for op in self.options)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_link(self)


class AbstractNode(AbstractBase, ABC):
    pass


class Union(AbstractBase):
    def __init__(
        self,
        name: str,
        types: t.List[str],
        *,
        description: t.Optional[str] = None,
    ):
        self.name = name
        self.types = types
        self.description = description

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.types
        )

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_union(self)


class Interface(AbstractBase):
    def __init__(
        self,
        name: str,
        fields: t.Sequence[t.Union["Field", "Link"]],
        *,
        description: t.Optional[str] = None,
    ):
        self.name = name
        self.fields = fields
        self.description = description

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.fields
        )

    @cached_property
    def fields_map(self) -> OrderedDict:
        return OrderedDict((f.name, f) for f in self.fields)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_interface(self)


class Input(AbstractBase):
    def __init__(
        self,
        name: str,
        arguments: t.Sequence[Option],
        *,
        description: t.Optional[str] = None,
    ):
        self.name = name
        self.arguments = arguments
        self.description = description
        # TODO(m.kind): support directives

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.arguments
        )

    @cached_property
    def arguments_map(self) -> OrderedDict[str, Option]:
        return OrderedDict((f.name, f) for f in self.arguments)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_input(self)


class Node(AbstractNode):
    """Collection of the fields and links, which describes some entity and
    relations with other entities

    Example::

        graph = Graph([
            Node('user', [
                Field('id', Integer, field_func),
                Field('name', String, field_func),
                Link('roles', Sequence[TypeRef['role']],
                     link_func, requires='id'),
            ]),
        ])

    """

    def __init__(
        self,
        name: t.Optional[str],
        fields: t.List[t.Union[Field, Link]],
        *,
        description: t.Optional[str] = None,
        directives: t.Optional[t.Sequence[SchemaDirective]] = None,
        implements: t.Optional[t.Sequence[str]] = None,
    ):
        """
        :param name: name of the node (None if Root node)
        :param fields: list of fields and links
        :param description: description of the node
        :param directives: list of directives for the node
        :param implements: list of interfaces implemented by the node
        """
        self.name = name
        self.fields = fields
        self.description = description
        self.directives: t.Tuple[SchemaDirective, ...] = tuple(directives or ())
        self.implements = tuple(implements or [])

    def __repr__(self) -> str:
        return "{}({!r}, {!r}, ...)".format(
            self.__class__.__name__, self.name, self.fields
        )

    @cached_property
    def fields_map(self) -> OrderedDict:
        return OrderedDict((f.name, f) for f in self.fields)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_node(self)

    def copy(self) -> "Node":
        return Node(
            name=self.name,
            fields=self.fields[:],
            description=self.description,
            directives=self.directives,
            implements=self.implements,
        )


class Root(Node):
    """Special implicit root node, starting point of the query execution

    Example::

        graph = Graph([
            Node('baz', [...]),
            Root([
                Field('foo', String, root_fields_func),
                Link('bar', Sequence[TypeRef['baz']],
                     to_baz_func, requires=None),
            ]),
        ])

    """

    def __init__(
        self,
        items: t.List[t.Union[Field, Link]],
    ):
        """
        :param items: list of fields, links and singleton nodes
        """
        super(Root, self).__init__(None, items)

    def __repr__(self) -> str:
        return "{}({!r}, {!r})".format(
            self.__class__.__name__, self.name, self.fields
        )

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_root(self)


class AbstractGraph(AbstractBase, ABC):
    pass


G = t.TypeVar("G", bound="Graph")


class Graph(AbstractGraph):
    """Collection of nodes - definition of the graph

    Example::

        graph = Graph([
            Node('foo', [...]),
            Node('bar', [...]),
            Root([...]),
        ])

    """

    def __init__(
        self,
        items: t.List[Node],
        data_types: t.Optional[t.Dict[str, t.Type[Record]]] = None,
        directives: t.Optional[t.Sequence[t.Type[SchemaDirective]]] = None,
        unions: t.Optional[t.List[Union]] = None,
        interfaces: t.Optional[t.List[Interface]] = None,
        enums: t.Optional[t.List[BaseEnum]] = None,
        scalars: t.Optional[t.List[t.Type[Scalar]]] = None,
        inputs: t.Optional[t.List[Input]] = None,
    ):
        """
        :param items: list of nodes
        """
        from .validate.graph import GraphValidator

        if unions is None:
            unions = []

        if interfaces is None:
            interfaces = []

        if enums is None:
            enums = []

        if scalars is None:
            scalars = []

        if inputs is None:
            inputs = []

        GraphValidator.validate(
            items, unions, interfaces, enums, scalars, inputs
        )

        self.items = GraphInit.init(items)
        self.unions = unions
        self.interfaces = interfaces
        self.interfaces_types = collect_interfaces_types(self.items, interfaces)
        self.enums: t.List[BaseEnum] = enums
        self.scalars = scalars
        self.inputs = inputs
        # TODO: deprecate data_types inputs
        self.data_types = data_types or {}
        self.__types__ = GraphTypes.get_types(
            self.items,
            self.unions,
            self.interfaces,
            self.enums,
            self.data_types,
        )
        self.directives: t.Tuple[t.Type[SchemaDirective], ...] = tuple(
            directives or ()
        )

    def __repr__(self) -> str:
        return "{}({!r})".format(self.__class__.__name__, self.items)

    def iter_root(self) -> t.Iterator[t.Union[Field, Link]]:
        """Iterate over nodes, and yield fields from all root nodes."""
        for node in self.items:
            if node.name is None:
                for field in node.fields:
                    yield field

    def iter_nodes(self) -> t.Iterator[Node]:
        for node in self.items:
            if node.name is not None:
                yield node

    @cached_property
    def root(self) -> Root:
        return Root(list(self.iter_root()))

    @cached_property
    def nodes(self) -> t.List[Node]:
        return list(self.iter_nodes())

    @cached_property
    def nodes_map(self) -> OrderedDict:
        return OrderedDict((e.name, e) for e in self.iter_nodes())

    @cached_property
    def unions_map(self) -> OrderedDict:
        return OrderedDict((u.name, u) for u in self.unions)

    @cached_property
    def interfaces_map(self) -> OrderedDict:
        return OrderedDict((i.name, i) for i in self.interfaces)

    @cached_property
    def enums_map(self) -> "OrderedDict[str, BaseEnum]":
        return OrderedDict((e.name, e) for e in self.enums)

    @cached_property
    def scalars_map(self) -> "OrderedDict[str, t.Type[Scalar]]":
        return OrderedDict((s.__type_name__, s) for s in self.scalars)

    @cached_property
    def inputs_map(self) -> "OrderedDict[str, Input]":
        return OrderedDict((i.name, i) for i in self.inputs)

    def accept(self, visitor: "AbstractGraphVisitor") -> t.Any:
        return visitor.visit_graph(self)

    @classmethod
    def from_graph(cls: t.Type[G], other: G, root: Root) -> G:
        """Create graph from other graph, with new root node.
        Useful for creating mutation graph from query graph.

        Example:
            MUTATION_GRAPH = Graph.from_graph(QUERY_GRAPH, Root([...]))
        """
        return cls(
            other.nodes + [root],
            data_types=other.data_types,
            directives=other.directives,
            unions=other.unions,
            interfaces=other.interfaces,
            enums=other.enums,
            scalars=other.scalars,
            inputs=other.inputs,
        )


class AbstractGraphVisitor(ABC):
    @abstractmethod
    def visit(self, obj: t.Any) -> t.Any:
        pass

    @abstractmethod
    def visit_option(self, obj: Option) -> t.Any:
        pass

    @abstractmethod
    def visit_field(self, obj: Field) -> t.Any:
        pass

    @abstractmethod
    def visit_link(self, obj: Link) -> t.Any:
        pass

    @abstractmethod
    def visit_node(self, obj: Node) -> t.Any:
        pass

    @abstractmethod
    def visit_union(self, obj: Union) -> t.Any:
        pass

    @abstractmethod
    def visit_interface(self, obj: Interface) -> t.Any:
        pass

    @abstractmethod
    def visit_input(self, obj: Input) -> t.Any:
        pass

    @abstractmethod
    def visit_enum(self, obj: BaseEnum) -> t.Any:
        pass

    @abstractmethod
    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Any:
        pass

    @abstractmethod
    def visit_root(self, obj: Root) -> t.Any:
        pass

    @abstractmethod
    def visit_graph(self, obj: Graph) -> t.Any:
        pass


class GraphVisitor(AbstractGraphVisitor):
    def visit(self, obj: t.Any) -> t.Any:
        return obj.accept(self)

    def visit_union(self, obj: "Union") -> t.Any:
        pass

    def visit_interface(self, obj: "Interface") -> t.Any:
        pass

    def visit_enum(self, obj: "BaseEnum") -> t.Any:
        pass

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Any:
        pass

    def visit_option(self, obj: "Option") -> t.Any:
        pass

    def visit_field(self, obj: Field) -> t.Any:
        for option in obj.options:
            self.visit(option)

    def visit_link(self, obj: Link) -> t.Any:
        for option in obj.options:
            self.visit(option)

    def visit_input(self, obj: Input) -> t.Any:
        for argument in obj.arguments:
            self.visit(argument)

    def visit_node(self, obj: Node) -> t.Any:
        for item in obj.fields:
            self.visit(item)

    def visit_root(self, obj: Root) -> t.Any:
        for item in obj.fields:
            self.visit(item)

    def visit_graph(self, obj: Graph) -> t.Any:
        for item in obj.items:
            self.visit(item)


class GraphTransformer(AbstractGraphVisitor):
    def visit(self, obj: t.Any) -> t.Any:
        return obj.accept(self)

    def visit_option(self, obj: Option) -> Option:
        return Option(
            obj.name, obj.type, default=obj.default, description=obj.description
        )

    def visit_field(self, obj: Field) -> Field:
        return Field(
            obj.name,
            obj.type,
            obj.func,
            options=[self.visit(op) for op in obj.options],
            description=obj.description,
            directives=obj.directives,
        )

    def visit_link(self, obj: Link) -> Link:
        return Link(
            obj.name,
            obj.type,
            obj.func,
            requires=obj.requires,
            options=[self.visit(op) for op in obj.options],
            description=obj.description,
            directives=obj.directives,
        )

    def visit_node(self, obj: Node) -> Node:
        return Node(
            obj.name,
            [self.visit(f) for f in obj.fields],
            description=obj.description,
            directives=obj.directives,
            implements=obj.implements,
        )

    def visit_input(self, obj: Input) -> Input:
        return Input(
            obj.name,
            [self.visit(a) for a in obj.arguments],
            description=obj.description,
        )

    def visit_union(self, obj: Union) -> Union:
        return Union(
            obj.name,
            obj.types,
            description=obj.description,
        )

    def visit_interface(self, obj: Interface) -> Interface:
        return Interface(
            obj.name,
            obj.fields,
            description=obj.description,
        )

    def visit_enum(self, obj: BaseEnum) -> BaseEnum:
        return obj

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Type[Scalar]:
        return obj

    def visit_root(self, obj: Root) -> Root:
        return Root([self.visit(f) for f in obj.fields])

    def visit_graph(self, obj: Graph) -> Graph:
        return Graph(
            [self.visit(node) for node in obj.items],
            obj.data_types,
            obj.directives,
            obj.unions,
            obj.interfaces,
            obj.enums,
            obj.scalars,
            obj.inputs,
        )


def apply(graph: G, transformers: List[GraphTransformer]) -> G:
    """Helper function to apply graph transformations

    Example:

    .. code-block:: python

        graph = hiku.graph.apply(graph, [AsyncGraphQLIntrospection()])

    """
    return reduce(lambda g, tr: tr.visit(g), transformers, graph)


class GraphInit(GraphTransformer):
    @classmethod
    def init(cls, items: t.List[Node]) -> t.List[Node]:
        self = cls()
        return [self.visit(i) for i in items]

    def visit_field(self, obj: Field) -> Field:
        field = super(GraphInit, self).visit_field(obj)
        postprocess = getattr(field.func, "__postprocess__", None)
        if postprocess is not None:
            postprocess(field)
        return field

    def visit_link(self, obj: Link) -> Link:
        link = super(GraphInit, self).visit_link(obj)
        postprocess = getattr(link.func, "__postprocess__", None)
        if postprocess is not None:
            postprocess(link)
        return link


class GraphTypes(GraphVisitor):
    def _visit_graph(
        self,
        items: t.List[Node],
        unions: t.List[Union],
        interfaces: t.List[Interface],
        enums: t.List[BaseEnum],
        data_types: t.Dict[str, t.Type[Record]],
    ) -> t.Dict[str, t.Type[Record]]:
        types = OrderedDict(data_types)
        roots = []
        for item in items:
            if item.name is not None:
                types[item.name] = self.visit(item)
            else:
                roots.append(self.visit(item))

        for union in unions:
            types[union.name] = self.visit(union)

        for interface in interfaces:
            types[interface.name] = self.visit(interface)

        for enum in enums:
            types[enum.name] = self.visit(enum)

        types["__root__"] = Record[
            chain.from_iterable(r.__field_types__.items() for r in roots)
        ]
        return types

    @classmethod
    def get_types(
        cls,
        items: t.List[Node],
        unions: t.List[Union],
        interfaces: t.List[Interface],
        enums: t.List[BaseEnum],
        data_types: t.Dict[str, t.Type[Record]],
    ) -> t.Dict[str, t.Type[Record]]:
        return cls()._visit_graph(items, unions, interfaces, enums, data_types)

    def visit_graph(self, obj: Graph) -> t.Dict[str, t.Type[Record]]:
        return self._visit_graph(
            obj.items, obj.unions, obj.interfaces, obj.enums, obj.data_types
        )

    def visit_node(self, obj: Node) -> t.Type[Record]:
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_root(self, obj: Root) -> t.Type[Record]:
        return Record[[(f.name, self.visit(f)) for f in obj.fields]]

    def visit_link(
        self, obj: Link
    ) -> t.Union[t.Type[TypeRef], t.Type[Optional], t.Type[Sequence]]:
        return obj.type

    def visit_field(self, obj: Field) -> t.Union[FieldT, AnyMeta]:
        return obj.type or Any

    def visit_union(self, obj: Union) -> Union:
        return obj

    def visit_enum(self, obj: BaseEnum) -> BaseEnum:
        return obj

    def visit_interface(self, obj: Interface) -> Interface:
        return obj

    def visit_scalar(self, obj: t.Type[Scalar]) -> t.Type[Scalar]:
        return obj
